@@ -14,9 +14,7 @@
 
 import getpass
 import re
-from typing import Optional
 
-import attr
 import click
 
 import releasetool.circleci
@@ -24,19 +22,15 @@ import releasetool.git
 import releasetool.github
 import releasetool.secrets
 import releasetool.commands.common
+from releasetool.commands.common import TagContext
+
+RELEASE_NOTES_HEADER_PATTERNS = [
+    r"This pull request was generated using releasetool\.\s+(.*)",
+    r":robot: .*\n---(.*)\n---",
+]
 
 
-@attr.s(auto_attribs=True, slots=True)
-class Context(releasetool.commands.common.GitHubContext):
-    package_name: Optional[str] = None
-    release_pr: Optional[dict] = None
-    release_tag: Optional[str] = None
-    release_version: Optional[str] = None
-    release_notes: Optional[str] = None
-    github_release: Optional[dict] = None
-
-
-def determine_release_pr(ctx: Context) -> None:
+def determine_release_pr(ctx: TagContext) -> None:
     click.secho(
         "> Let's figure out which pull request corresponds to your release.", fg="cyan"
     )
@@ -53,14 +47,17 @@ def determine_release_pr(ctx: Context) -> None:
     ctx.release_pr = pulls[pull_idx - 1]
 
 
-def determine_release_tag(ctx: Context) -> None:
+def _parse_release_tag(head_ref: str) -> str:
+    match = re.match(r"release.*-(v\d+\.\d+\.\d+)", head_ref)
+    return match.group(1) if match else None
+
+
+def determine_release_tag(ctx: TagContext) -> None:
     click.secho("> Determining what the release tag should be.", fg="cyan")
     head_ref = ctx.release_pr["head"]["ref"]
-    match = re.match("release-.+-(v\d+\.\d+\.\d+)", head_ref)
+    ctx.release_tag = _parse_release_tag(head_ref)
 
-    if match is not None:
-        ctx.release_tag = match.group(1)
-    else:
+    if ctx.release_tag is None:
         click.secho(
             "I couldn't determine what the release tag should be from the PR's"
             f"head ref {head_ref}.",
@@ -73,36 +70,35 @@ def determine_release_tag(ctx: Context) -> None:
     click.secho(f"Release tag is {ctx.release_tag}.")
 
 
-def determine_package_name_and_version(ctx: Context) -> None:
+def determine_package_name_and_version(ctx: TagContext) -> None:
     click.secho("> Determining the release version.", fg="cyan")
-    match = re.match("v(\d+\.\d+\.\d+)", ctx.release_tag)
+    match = re.match(r"v(\d+\.\d+\.\d+)", ctx.release_tag)
     ctx.release_version = match.group(1)
     click.secho(f"Package version: {ctx.release_version}.")
 
 
-def get_release_notes(ctx: Context) -> None:
+def _parse_release_notes(pr_body) -> str:
+    for pattern in RELEASE_NOTES_HEADER_PATTERNS:
+        match = re.search(pattern, pr_body, re.DOTALL | re.MULTILINE)
+        if match is not None:
+            return match.group(1).strip()
+
+    return ""
+
+
+def get_release_notes(ctx: TagContext) -> None:
     click.secho("> Grabbing the release notes.", fg="cyan")
-
-    match = re.search(
-        r"This pull request was generated using releasetool\.\s+(.*)",
-        ctx.release_pr["body"],
-        re.DOTALL | re.MULTILINE,
-    )
-    if match is not None:
-        ctx.release_notes = match.group(1).strip()
-    else:
-        ctx.release_notes = ""
-
+    ctx.release_notes = _parse_release_notes(ctx.release_pr["body"])
     click.secho(f"Release notes:\n{ctx.release_notes}")
 
 
-def create_release(ctx: Context) -> None:
+def create_release(ctx: TagContext) -> None:
     click.secho("> Creating the release.")
 
     ctx.github_release = ctx.github.create_release(
         repository=ctx.upstream_repo,
         tag_name=ctx.release_tag,
-        target_committish=ctx.release_pr["merge_commit_sha"],
+        target_commitish=ctx.release_pr["merge_commit_sha"],
         name=ctx.release_tag,
         body=ctx.release_notes,
     )
@@ -114,19 +110,41 @@ def create_release(ctx: Context) -> None:
         ctx.upstream_repo, ctx.release_pr["number"], release_location_string
     )
 
+    repo_short_name = ctx.upstream_repo.split("/")[-1]
+    ctx.kokoro_job_name = (
+        f"cloud-devrel/client-libraries/java/{repo_short_name}/release/stage"
+    )
+    ctx.github.update_pull_labels(
+        ctx.release_pr, add=["autorelease: tagged"], remove=["autorelease: pending"]
+    )
+    releasetool.commands.common.publish_via_kokoro(ctx)
 
-def tag() -> None:
-    ctx = Context()
 
-    click.secho(f"o/ Hey, {getpass.getuser()}, let's tag a release!", fg="magenta")
+def tag(ctx: TagContext = None) -> TagContext:
+    if not ctx:
+        ctx = TagContext()
 
-    releasetool.commands.common.setup_github_context(ctx)
+    if ctx.interactive:
+        click.secho(f"o/ Hey, {getpass.getuser()}, let's tag a release!", fg="magenta")
+
+    if ctx.github is None:
+        releasetool.commands.common.setup_github_context(ctx)
+
+    if ctx.release_pr is None:
+        determine_release_pr(ctx)
 
-    determine_release_pr(ctx)
     determine_release_tag(ctx)
     determine_package_name_and_version(ctx)
-    get_release_notes(ctx)
 
+    # If the release already exists, don't do anything
+    if releasetool.commands.common.release_exists(ctx):
+        click.secho(f"{ctx.release_tag} already exists.", fg="magenta")
+        return ctx
+
+    get_release_notes(ctx)
     create_release(ctx)
 
-    click.secho(f"\o/ All done!", fg="magenta")
+    if ctx.interactive:
+        click.secho("\\o/ All done!", fg="magenta")
+
+    return ctx
